/*
 *                            *
 *                           * *
 *                          *   *
 *                     ***************
 *                      * *       * *
 *                       *  MUMPS  *
 *                      * *       * *
 *                     ***************
 *                          *   *
 *                           * *
 *                            *
 *
 *   global_bdb.c
 *    BerkeleyDB global handler
 *
 *  
 *   Author: John P. Willis <jpw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2020 Coherent Logic Development LLC
 *
 *
 *   This file is part of FreeM.
 *
 *   FreeM is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Affero Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   FreeM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Affero Public License for more details.
 *
 *   You should have received a copy of the GNU Affero Public License
 *   along with FreeM.  If not, see <https://www.gnu.org/licenses/>.
 *
 **/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"

#if defined(GVH_BDB)

#include <db.h>

#include "mpsdef.h"
#include "mref.h"
#include "iniconf.h"

#define PATH_MAX 2048

typedef struct frm_bdb_gbl_session {

    char name[256];
    DB *database;

    unsigned long int n_gets;
    unsigned long int n_sets;
    unsigned long int n_kills;
    unsigned long int n_orders;
    unsigned long int n_queries;
    unsigned long int n_datas;

    unsigned long int n_writes_until_flush;
    unsigned long int n_flushes;
    
    struct frm_bdb_gbl_session *next;

} frm_bdb_gbl_session;

frm_bdb_gbl_session *frm_bdb_gbl_sessions_head;

frm_bdb_gbl_session *frm_bdb_gbl_get_session(char *);
void frm_bdb_gbl_get_filename(char *, char *);
void frm_bdb_gbl_get(char *, char *, char *);
void frm_bdb_gbl_set(char *, char *, char *);
void frm_bdb_gbl_order(char *, char *, char *);
void frm_bdb_gbl_data(char *, char *, char *);
void frm_bdb_gbl_query(char *, char *, char *);
void frm_bdb_gbl_revquery(char *name, char *key, char *data);
void frm_bdb_chkflush(frm_bdb_gbl_session *);
unsigned long int frm_bdb_char_count(char *key, char c);
    
void frm_bdb_chkflush(frm_bdb_gbl_session *session)
{

    if (--(session->n_writes_until_flush) == 0) {
        session->database->sync (session->database, 0);        
        session->n_writes_until_flush = bdb_flush_threshold;
        session->n_flushes++;
    }
        
}

void frm_bdb_flush_all(void)
{
    frm_bdb_gbl_session *t;

    for (t = frm_bdb_gbl_sessions_head; t != NULL; t = t->next) {
        t->n_writes_until_flush = 1;
        frm_bdb_chkflush (t);
    }        
}

void frm_bdb_close_all(void)
{
    frm_bdb_gbl_session *t;

    for (t = frm_bdb_gbl_sessions_head; t != NULL; t = t->next) {
        t->n_writes_until_flush = 1;
        frm_bdb_chkflush (t);
        t->database->close (t->database, 0);
    }
}

frm_bdb_gbl_session *frm_bdb_gbl_get_session(char *name)
{    
    frm_bdb_gbl_session *t;
    
    char *db_file = (char *) malloc (PATH_MAX * sizeof (char));
    NULLPTRCHK(db_file,"frm_bdb_gbl_get_session");

    frm_bdb_gbl_get_filename (db_file, name);
    
    for (t = frm_bdb_gbl_sessions_head; t != NULL; t = t->next) {
        
        if (strcmp (t->name, name) == 0) {
            /* found an existing session for this global */
            return t;            
        }

    }

    /* no sessions yet exist. we make new one. */
    t = (frm_bdb_gbl_session *) malloc (sizeof (frm_bdb_gbl_session));
    NULLPTRCHK(t,"frm_bdb_gbl_session");

    /* set up the t->database struct */
    db_create (&(t->database), NULL, 0);
    
    /* next, we open the database */
    t->database->open (t->database, NULL, db_file, NULL, DB_BTREE, DB_CREATE, 0);

    strcpy (t->name, name);

    t->n_gets = 0;
    t->n_sets = 0;
    t->n_kills = 0;
    t->n_orders = 0;
    t->n_queries = 0;
    t->n_datas = 0;
    t->n_writes_until_flush = bdb_flush_threshold;
    t->n_flushes = 0;
    
    t->next = frm_bdb_gbl_sessions_head;
    frm_bdb_gbl_sessions_head = t;

     return t;
}

void frm_bdb_gbl_get_filename(char *filename, char *name)
{
    char t[STRLEN];
    char *p;

    if (name[1] == '%' || name[1] == '$') {
        p = gloplib;
    }
    else {
        p = glopath;
    }

    stcpy (t, p);
    stcnv_m2c (t);
    
    snprintf (filename, PATH_MAX, "%s/%s.freem", t, name);    
    
}


void global_bdb (short action, char *key, char *data)
{
    char name[256];
    char *ptr = key;
    register char ch;
    register int i;
    
    i = 0;
    while ((ch = *(ptr++)) != DELIM && ch != EOL && i < 256) {
        name[i++] = ch;
    }

    name[i] = '\0';
    
    switch (action) {

        case get_sym:

            frm_bdb_gbl_get (name, key, data);
            break;
            
        case set_sym:

            frm_bdb_gbl_set (name, key, data);
            break;

        case fra_query:

            frm_bdb_gbl_query (name, key, data);
            break;

        case fra_order:
            
            frm_bdb_gbl_order (name, key, data);
            break;

        case dat:
            frm_bdb_gbl_data (name, key, data);
            break;
            
        default:
            break;
            
    }

}

void frm_bdb_gbl_get(char *name, char *key, char *data)
{

    DBT bdb_key;
    DBT bdb_data;
    int ret;
    
    frm_bdb_gbl_session *session = frm_bdb_gbl_get_session (name); 

    memset (&bdb_key, 0, sizeof (DBT));
    memset (&bdb_data, 0, sizeof (DBT));

    bdb_key.data = key;
    bdb_key.size = stlen (key) + 1;
    bdb_data.flags = DB_DBT_MALLOC;

    ret = session->database->get(session->database, NULL, &bdb_key, &bdb_data, 0);

    switch (ret) {

        case 0:

            session->n_gets++;
            
            stcpy (data, (char *) bdb_data.data);
            ierr = OK;
            break;

        case DB_NOTFOUND:

            *data = '\201';            
            ierr = M7;
            break;

    }
    
}

void frm_bdb_gbl_set(char *name, char *key, char *data)
{

    DBT bdb_key;
    DBT bdb_data;
    int ret;
    
    frm_bdb_gbl_session *session = frm_bdb_gbl_get_session (name); 

    memset (&bdb_key, 0, sizeof (DBT));
    memset (&bdb_data, 0, sizeof (DBT));

    bdb_key.data = key;
    bdb_key.size = stlen (key) + 1;

    bdb_data.data = data;
    bdb_data.size = stlen (data) + 1;
    
    ret = session->database->put (session->database, NULL, &bdb_key, &bdb_data, 0);

    if (ret == 0) {
        session->n_sets++;
        frm_bdb_chkflush (session);
        ierr = OK;
    }
    
}

void frm_bdb_gbl_order(char *name, char *key, char *data)
{
    DBC *cur;
    DBT bdb_key;
    DBT bdb_data;
    int ret;
    int initial_subct;
    
    frm_bdb_gbl_session *session = frm_bdb_gbl_get_session (name); 
    freem_ref_t *r = (freem_ref_t *) malloc (sizeof (freem_ref_t));

    initial_subct = frm_bdb_char_count (key, DELIM);
   
    memset (&bdb_key, 0, sizeof (DBT));
    memset (&bdb_data, 0, sizeof (DBT));
    
    bdb_key.data = key;
    bdb_key.size = stlen (key) + 1;

    ret = session->database->cursor (session->database, NULL, &cur, 0);
    
    cur->c_get (cur, &bdb_key, &bdb_data, DB_SET);

    session->n_orders++;
    
    for (;;) {

        cur->c_get (cur, &bdb_key, &bdb_data, DB_NEXT);        
            
        if (frm_bdb_char_count (bdb_key.data, DELIM) == initial_subct) {

            mref_init (r, MREF_RT_GLOBAL, name);
            internal_to_mref (r, bdb_key.data);

            stcpy (data, mref_get_subscript (r, r->subscript_count - 1));
            stcnv_c2m (data);
            
            goto order_done;
        }

    }

    *data = '\201';
    

order_done:    
    ret = cur->c_close (cur);
    free (r);
}

void frm_bdb_gbl_data(char *name, char *key, char *data)
{
    DBC *cur;
    DBT bdb_key;
    DBT bdb_data;
    int ret;

    char *k_buf = (char *) malloc (STRLEN * sizeof (char));
    char *d_buf = (char *) malloc (STRLEN * sizeof (char));
    char *t_buf = (char *) malloc (STRLEN * sizeof (char));
    
    short gv_defined = FALSE;
    short gv_has_data = FALSE;
    short gv_has_subscripts = FALSE;
    
    frm_bdb_gbl_session *session = frm_bdb_gbl_get_session (name);
    freem_ref_t *r = (freem_ref_t *) malloc (sizeof (freem_ref_t));

    memset (&bdb_key, 0, sizeof (DBT));
    memset (&bdb_data, 0, sizeof (DBT));

    /* determine whether or not the global is defined */
    frm_bdb_gbl_get (name, key, d_buf);
    if (ierr != M7) gv_defined = TRUE;    

    if (stlen (d_buf)) gv_has_data = TRUE;
    
    bdb_key.data = key;
    bdb_key.size = stlen (key) + 1;

data_done:
    free (k_buf);
    free (d_buf);
    free (t_buf);

    return;
                     
}

void frm_bdb_gbl_query(char *name, char *key, char *data)
{
    DBC *cur;
    DBT bdb_key;
    DBT bdb_data;
    int ret;

    frm_bdb_gbl_session *session = frm_bdb_gbl_get_session (name); 
    freem_ref_t *r = (freem_ref_t *) malloc (sizeof (freem_ref_t));
    
    memset (&bdb_key, 0, sizeof (DBT));
    memset (&bdb_data, 0, sizeof (DBT));

    bdb_key.data = key;
    bdb_key.size = stlen (key) + 1;

    ret = session->database->cursor (session->database, NULL, &cur, 0);

    cur->c_get (cur, &bdb_key, &bdb_data, DB_SET);
    
    if (cur->c_get (cur, &bdb_key, &bdb_data, DB_NEXT) == 0) {        

        key = bdb_key.data;       

        session->n_queries++;
        
        mref_init (r, MREF_RT_GLOBAL, "");
        internal_to_mref (r, key);
        mref_to_external (r, data);
        stcnv_c2m (data);
        
    }
    else {
        *data = '\201';
    }

    ret = cur->c_close (cur);
    free (r);    
}

void frm_bdb_gbl_revquery(char *name, char *key, char *data)
{
    DBC *cur;
    DBT bdb_key;
    DBT bdb_data;
    int ret;

    frm_bdb_gbl_session *session = frm_bdb_gbl_get_session (name); 
    freem_ref_t *r = (freem_ref_t *) malloc (sizeof (freem_ref_t));
    
    memset (&bdb_key, 0, sizeof (DBT));
    memset (&bdb_data, 0, sizeof (DBT));

    bdb_key.data = key;
    bdb_key.size = stlen (key) + 1;

    ret = session->database->cursor (session->database, NULL, &cur, 0);

    cur->c_get (cur, &bdb_key, &bdb_data, DB_SET);
    
    if (cur->c_get (cur, &bdb_key, &bdb_data, DB_PREV) == 0) {        

        key = bdb_key.data;       

        session->n_queries++;
        
        mref_init (r, MREF_RT_GLOBAL, "");
        internal_to_mref (r, key);
        mref_to_external (r, data);
        stcnv_c2m (data);
        
    }
    else {
        *data = '\201';
    }

    ret = cur->c_close (cur);
    free (r);
    
}

unsigned long int frm_bdb_char_count(char *key, char c)
{

    unsigned long int i;
    unsigned long int ct;
    char ch;

    ct = 0;
    i = 0;
    while ((ch = key[i++]) != EOL) if (ch == c) ct++;

    return ct;

}

void frm_bdb_gbl_session_dump(void)
{
    frm_bdb_gbl_session *s;

    printf ("\r\n%-20s%-15s%-15s%-15s%-15s%-15s%-15s%-10s%-10s\r\n", "GLOBAL","GETCT","SETCT","KILLCT","$ORDERCT","$QUERYCT","$DATACT","FLUSH IN","FLUSHCT");
    printf ("%-20s%-15s%-15s%-15s%-15s%-15s%-15s%-10s%-10s\r\n", "------","-----","-----","------","--------","--------","-------","--------","-------");
    
    
    for (s = frm_bdb_gbl_sessions_head; s != NULL; s = s->next) {        
        printf("%-20s%-15ld%-15ld%-15ld%-15ld%-15ld%-15ld%-10ld%-10ld\r\n", s->name, s->n_gets, s->n_sets, s->n_kills, s->n_orders, s->n_queries, s->n_datas, s->n_writes_until_flush, s->n_flushes);        
    }

    printf ("\r\nENGINE [%-20s]:\t%s\r\n", "SYSTEM", gbl_s_engine);
    printf ("ENGINE [%-20s]:\t%s\r\n", nsname, gbl_u_engine);
    printf ("CACHE SYNC THRESHOLD         :\t%ld\r\n\r\n", bdb_flush_threshold);    
}
#endif
